import hashlib
import urllib.parse
from xnatutils.base import (
    sanitize_re,
    get_resource_name,
    session_modality_re,
    connect,
)
from xnatutils.exceptions import (
    XnatUtilsUsageError,
    XnatUtilsError,
    XnatUtilsDigestCheckFailedError,
    XnatUtilsNoMatchingSessionsException,
)

HASH_CHUNK_SIZE = 2**20

BAD_CHAR_MSG = "(must only contain alpha-numeric characters and underscores)"


def resource(session, scan, *filenames, **kwargs):
    """
    Uploads datasets to an XNAT instance project (requires manager privileges for the
    project).

    The format of the uploaded file is guessed from the file extension
    (recognised extensions are '.nii', '.nii.gz', '.mif'), the scan entry is
    created in the session and if 'create_session' kwarg is True the
    subject and session are created if they are not already present, e.g.

        >>> xnatutils.put('TEST001_001_MR01', 'a_dataset', ['test.nii.gz'],
                          create_session=True)

    NB: If the scan already exists the 'overwrite' kwarg must be provided to
    overwrite it.

    User credentials can be stored in a ~/.netrc file so that they don't need
    to be entered each time a command is run. If a new user provided or netrc
    doesn't exist the tool will ask whether to create a ~/.netrc file with the
    given credentials.

    Parameters
    ----------
    session : str
        Name of the session to upload the dataset to
    scan : str
        Name for the dataset on XNAT
    filenames : list(str)
        Filenames of the dataset(s) to upload to XNAT or a directory containing
        the datasets.
    overwrite : bool
        Allow overwrite of existing dataset
    create_session : bool
        Create the required session on XNAT to upload the the dataset to
    resource_name : str
        The name of the resource (the data format) to
        upload the dataset to. If not provided the format
        will be determined from the file extension (i.e.
        in most cases it won't be necessary to specify
    project_id : str
        The ID of the project to upload the dataset to
    subject_id : str
        The ID of the subject to upload the dataset to
    scan_id : str
        The ID for the scan (defaults to the scan type)
    date : str
        The date of the scan
    modality : str
        The modality of the session to upload
    user : str
        The user to connect to the server with
    loglevel : str
        The logging level to display. In order of increasing verbosity
        ERROR, WARNING, INFO, DEBUG.
    connection : xnat.Session
        An existing XnatPy session that is to be reused instead of
        creating a new session. The session is wrapped in a dummy class
        that disables the disconnection on exit, to allow the method to
        be nested in a wider connection context (i.e. reuse the same
        connection between commands).
    server : str | int | None
        URI of the XNAT server to connect to. If not provided connect
        will look inside the ~/.netrc file to get a list of saved
        servers. If there is more than one, then they can be selected
        by passing an index corresponding to the order they are listed
        in the .netrc
    use_netrc : bool
        Whether to load and save user credentials from netrc file
        located at $HOME/.netrc
    """
    # Set defaults for kwargs
    create_session = kwargs.pop(
        "create_session",
        False,
    )
    resource_name = kwargs.pop("resource_name", None)
    project_id = kwargs.pop("project_id", None)
    subject_id = kwargs.pop("subject_id", None)
    date = kwargs.pop("date", None)
    scan_id = kwargs.pop("scan_id", None)
    modality = kwargs.pop("modality", None)
    if sanitize_re.match(session):
        raise XnatUtilsUsageError(
            f"Session '{session}' is not a valid session name - {BAD_CHAR_MSG}"
        )
    if sanitize_re.match(scan):
        raise XnatUtilsUsageError(
            f"Scan {scan} is not a valid dataset name - {BAD_CHAR_MSG}"
        )
    scan = urllib.parse.quote(scan)

    if resource_name is None:
        if len(filenames) == 1:
            resource_name = get_resource_name(filenames[0])
        else:
            raise XnatUtilsUsageError(
                "'resource_name' option needs to be provided when uploading "
                "multiple files"
            )
    else:
        resource_name = resource_name.upper()
    with connect(**kwargs) as login:
        if modality is None:
            match = session_modality_re.match(session)
            if match is None:
                modality = "MR"  # The default
            else:
                modality = match.group(1)
        session_cls, scan_cls = get_xnat_classes(login, modality)
        try:
            xsession = login.experiments[session]
        except KeyError:
            if create_session:
                if project_id is None and subject_id is None:
                    try:
                        project_id, subject_id, _ = session.split("_")
                    except ValueError:
                        raise XnatUtilsUsageError(
                            "Must explicitly provide project and subject IDs "
                            "if session ID ({}) scheme doesn't match "
                            "<project>_<subject>_<visit> convention, i.e. "
                            "have exactly 2 underscores".format(session)
                        )
                if project_id is None:
                    project_id = session.split("_")[0]
                if subject_id is None:
                    subject_id = "_".join(session.split("_")[:2])
                try:
                    xproject = login.projects[project_id]
                except KeyError:
                    raise XnatUtilsUsageError(
                        "Cannot create session '{}' as '{}' does not exist "
                        "(or you don't have access to it)".format(session, project_id)
                    )
                # Creates a corresponding subject and session if they don't
                # exist
                xsubject = login.classes.SubjectData(label=subject_id, parent=xproject)
                xsession = session_cls(label=session, parent=xsubject)
                if date is not None:
                    xsession.date = date
            else:
                raise XnatUtilsNoMatchingSessionsException(
                    "'{}' session does not exist, to automatically create it "
                    "please use '--create_session' option.".format(session)
                )
        xdataset = scan_cls(
            id=(scan_id if scan_id is not None else scan),
            type=scan,
            parent=xsession,
        )
        resource = None
        # get the existing resource, if there is one
        # this means that we can reconnect to a resource and add files to it
        # after being interrupted
        try:
            resource = xdataset.resources[resource_name]
        except KeyError:
            resource = xdataset.create_resource(resource_name)
        return resource


def get_xnat_classes(login, modality):
    """
    Tries to deduce the XNAT session and scan classes from the modality.
    If it doesn't work, falls back to MrSessionData and MrScanData
    --
    modality: str like "MR", "CT", "PET", etc
    returns: classes of session, scan
    """
    if modality == "MRPT":
        session_cls = login.classes.PetmrSessionData
        scan_cls = login.classes.MrScanData
    else:
        mcap = modality.capitalize()
        try:
            session_cls = getattr(login.classes, mcap + "SessionData")
            scan_cls = getattr(login.classes, mcap + "ScanData")
        except AttributeError:
            try:
                session_cls = login.classes.MrSessionData
                scan_cls = login.classes.MrScanData
            except AttributeError:
                # Old name < 1.8
                session_cls = login.clases.mrSessionData
                scan_cls = login.classes.mrScanData
    return session_cls, scan_cls


def calculate_checksum(fname):
    try:
        file_hash = hashlib.md5()
        with open(fname, "rb") as f:
            for chunk in iter(lambda: f.read(HASH_CHUNK_SIZE), b""):
                file_hash.update(chunk)
        return file_hash.hexdigest()
    except OSError:
        raise XnatUtilsDigestCheckFailedError(
            "Could not check digest of '{}' ".format(fname)
        )


def get_digests(resource):
    """
    Downloads the MD5 digests associated with the files in a resource.
    These are saved with the downloaded files in the cache and used to
    check if the files have been updated on the server
    """
    result = resource.xnat_session.get(resource.uri + "/files")
    if result.status_code != 200:
        raise XnatUtilsError(
            "Could not download metadata for resource {}. Files "
            "may have been uploaded but cannot check checksums".format(resource.id)
        )
    return dict((r["Name"], r["digest"]) for r in result.json()["ResultSet"]["Result"])



import logging
import json
from openpyxl import load_workbook
from pathlib import Path

from xnatuploader.matcher import Matcher
from xnatuploader.xnatuploader import scan, collate_uploads
from xnatuploader.workbook import load_config, new_workbook

logger = logging.getLogger(__name__)


def assert_worksheets_equal(expect, got):
    grows = list(got.values)
    for row in expect.values:
        assert grows[0] == row
        grows = grows[1:]
    assert len(grows) == 0


def test_scan(tmp_path, test_files):
    fileset = test_files["basic"]
    config = load_config(fileset["config_excel"])
    matcher = Matcher(config)
    scanned = tmp_path / "scanned.xlsx"
    new_workbook(scanned)
    scan(matcher, Path(fileset["dir"]), scanned, include_unmatched=True)
    expect_wb = load_workbook(fileset["scanned_excel"])
    got_wb = load_workbook(scanned)
    assert "Files" in got_wb
    assert_worksheets_equal(expect_wb["Files"], got_wb["Files"])


def test_collation(tmp_path, test_files, uploads_dict):
    fileset = test_files["basic"]
    config = load_config(fileset["config_excel"])
    matcher = Matcher(config)
    log = tmp_path / "log.xlsx"
    new_workbook(log)
    scan(matcher, Path(fileset["dir"]), log)
    project_id = uploads_dict["project"]
    wb = load_workbook(log)
    ws = wb["Files"]
    header = True
    files = []
    for row in ws.values:
        if header:
            header = False
        else:
            matchfile = matcher.from_spreadsheet(row)
            files.append(matchfile)
    skipped, uploads = collate_uploads(project_id, files)
    for session_scan, upload in uploads.items():
        uploads[session_scan] = [f.file for f in uploads[session_scan].files]
    assert uploads == uploads_dict["uploads"]
    assert len(skipped) == uploads_dict["skipped"]


def test_sanitisation_collisions(tmp_path, test_files, sanitised_dict):
    fileset = test_files["sanitisation"]
    config_file = fileset["config"]
    with open(config_file, "r") as fh:
        config = json.load(fh)
    matcher = Matcher(config)
    logger.warning(f"Testing santisation in {fileset}")
    log = tmp_path / "log.xlsx"
    new_workbook(log)
    scan(matcher, Path(fileset["dir"]), log)
    project_id = sanitised_dict["project"]
    wb = load_workbook(log)
    ws = wb["Files"]
    header = True
    files = []
    for row in ws.values:
        if header:
            header = False
        else:
            matchfile = matcher.from_spreadsheet(row)
            files.append(matchfile)
    logger.warning(f"File list = {files}")
    skipped, uploads = collate_uploads(project_id, files)
    for session_scan, upload in uploads.items():
        uploads[session_scan] = [f.file for f in uploads[session_scan].files]
    assert uploads == sanitised_dict["uploads"]
    assert len(skipped) == sanitised_dict["skipped"]


def test_collation_skips(tmp_path, test_files, uploads_dict):
    fileset = test_files["basic"]
    config = load_config(fileset["config_excel"])
    matcher = Matcher(config)
    log = tmp_path / "log.xlsx"
    new_workbook(log)
    scan(matcher, Path(fileset["dir"]), log)
    project_id = uploads_dict["project"]
    wb = load_workbook(log)
    ws = wb["Files"]
    header = True
    files = []
    skip_n = 4  # mark the first four files as "success"
    n = 0
    for row in ws.values:
        if header:
            header = False
        else:
            matchfile = matcher.from_spreadsheet(row)
            if n < skip_n:
                if matchfile.selected:
                    matchfile.status = "success"
                    logger.warning(f"Skipping {matchfile.file}")
                    n += 1
            files.append(matchfile)
    skip, uploads = collate_uploads(project_id, files)
    assert len(skip) == uploads_dict["skipped"] + skip_n
    for session_scan, upload in uploads.items():
        uploads[session_scan] = [f.file for f in uploads[session_scan].files]
    # remove skipped files from the fixture dictionary
    upload_skipped = {}
    skipstr = [f.file for f in skip]
    for subject_id, files in uploads_dict["uploads"].items():
        upload_files = [f for f in files if f not in skipstr]
        # don't add if the list of files is empty
        if upload_files:
            upload_skipped[subject_id] = upload_files
    assert uploads == upload_skipped



from openpyxl import Workbook, load_workbook
from openpyxl.styles.alignment import Alignment

FILE_COLUMN_WIDTH = 50
HELP_COLUMN_WIDTH = 25
HELP_ROW_HEIGHT = 90

HELP_TEXT = """
Paths are a set of patterns to be matched against file paths in the source directory.

Values in curly brackets like {SubjectName} are captured into variables.
Variables with names in all caps like {YYYY} will only match numbers.
The special value * and ** match one or more than one intervening directory.

Mappings are used to combine the variables captured from a path to the
XNAT hierarchy Subject / Session / Dataset.

You can also use metadata from the DICOM files in the mappings, by
using the syntax "DICOM:MetadataName" - for example, DICOM:StudyDate

The XNAT section configures the XNAT project ID and server.
"""


class WorkbookError(Exception):
    pass


def new_workbook(file):
    """
    Make a new spreadsheet with the instructions worksheet and default
    patterns
    --
    file: filepath.Path
    """
    wb = Workbook()
    ws = wb.active
    for col in "ABCDE":
        ws.column_dimensions[col].width = HELP_COLUMN_WIDTH

    ws.title = "Configuration"
    ws["A1"] = "Instructions"
    ws["A2"] = HELP_TEXT
    ws["A2"].alignment = Alignment(wrapText=True)
    ws.merge_cells("A2:F2")
    ws.row_dimensions[2].height = HELP_ROW_HEIGHT
    ws["A4"] = "Configuration"
    ws["A5"] = "Paths"
    ws["B5"] = "DICOM"
    ws["C5"] = "{SubjectName}-{ID}"
    ws["D5"] = "**"
    ws["E5"] = "{Directory}"
    ws["F5"] = "{filename}.dcm"
    ws["A10"] = "Mappings"
    ws["B10"] = "Subject"
    ws["C10"] = "ID"
    ws["B11"] = "Session"
    ws["C11"] = "DICOM:StudyDate"
    ws["B12"] = "Dataset"
    ws["C12"] = "Directory"
    ws["A14"] = "XNAT"
    ws["B14"] = "Project"
    ws["C14"] = "Test001"
    ws["B15"] = "Server"
    ws["C15"] = "http://localhost:8080"
    wb.save(file)


def add_filesheet(wb, matcher, debug):
    """
    Adds a worksheet to a workbook with headers from the matcher object
    and niceties such as a wider column for the file names.

    If debug is true, an existing worksheet with the name "Files" is renamed
    to "Files-prev" - otherwise it's deleted.
    ---
    wb: a Workbook
    matcher: a Matcher
    debug: bool

    returns: the new worksheet
    """
    if "Files" in wb:
        if debug:
            old_files = wb["Files"]
            # older versions will be automatically renamed to Files-prev1,
            # Files-prev2 etc
            old_files.title = "Files-prev"
        else:
            wb.remove_sheet(wb["Files"])
    ws = wb.create_sheet("Files")
    ws.column_dimensions["B"].width = FILE_COLUMN_WIDTH
    ws.append(matcher.headers)
    return ws


def load_config(excelfile):
    wb = load_workbook(excelfile)
    if "Configuration" not in wb:
        raise WorkbookError(f"No worksheet named 'Configuration' in {excelfile}")
    ws = wb["Configuration"]
    config = {}
    section = None
    sections = ["paths", "mappings", "xnat"]
    for row in ws:
        if row[0].value is not None:
            section = row[0].value.lower()
        var = row[1].value
        if var is not None and section is not None:
            if section not in config:
                config[section] = {}
            cells = [cell.value for cell in row[2:] if cell.value is not None]
            if section == "xnat":
                config[section][var] = cells[0]
            else:
                config[section][var] = cells
    missing = [s for s in sections if s not in config]
    if len(missing) > 0:
        raise WorkbookError(f"Missing config sections: {missing}")
    return config



import re
import logging
from pydicom import dcmread
from pydicom.errors import InvalidDicomError

XNAT_HIERARCHY = ["Subject", "Session", "Dataset"]
DICOM_PARAMS = ["Modality", "StudyDescription", "StudyDate"]

logger = logging.getLogger(__name__)


class RecipeException(Exception):
    pass


class Matcher:
    """
    A Matcher is a set of recipes for matching against filepaths and mappings
    which map the captured value to XNAT parameters. It's used to perform
    the matching, returning FileMatch objects, and also has methods which are
    used to get headers for the spreadsheets (since these will vary according
    to the parameters captured in the recipes)
    """

    def __init__(self, config, loglevel="WARNING"):
        """
        config: dict with "paths" and "mappings"
        """
        logger.setLevel(loglevel)
        logch = logging.StreamHandler()
        logch.setLevel(loglevel.upper())
        logger.addHandler(logch)
        self.parse_recipes(config["paths"])
        self.mappings = config["mappings"]
        self._headers = None
        self._dicom_params = set(DICOM_PARAMS)
        for k, vs in self.mappings.items():
            for v in vs:
                if v[:5] == "DICOM":
                    self._dicom_params.add(v[6:])
                elif v not in self.params:
                    raise Exception(
                        f"Value {v} in mapping for {k} not defined in a path"
                    )
        if set(self.mappings.keys()) != set(XNAT_HIERARCHY):
            raise Exception(f"Must have mappings for each of {XNAT_HIERARCHY}")

    @property
    def headers(self):
        if self._headers is None:
            self._headers = [
                "Pattern",
                "File",
                "Filename",
                "Upload",
                "Status",
                "SessionLabel",
            ]
            self._headers += XNAT_HIERARCHY + self.dicom_params + self.params
        return self._headers

    @property
    def dicom_params(self):
        return sorted(self._dicom_params)

    def parse_recipes(self, recipe_config):
        """
        Parses a set of recipes from the config, keyed by label. The value for
        each label should be a list of recipe strings. Builds the params
        variable, which is a list of all the params, and a list of compiled
        regular expressions.
        ---
        recipe_config: dict of { str: list of str  }
        """
        self.params = []
        self.recipes = {}
        for label, patterns in recipe_config.items():
            self.recipes[label] = []
            for pattern in patterns:
                params, regexp = self.parse_recipe(label, pattern)
                self.recipes[label].append(regexp)
                self.params += [p for p in params if p not in self.params]

    def parse_recipe(self, label, recipe):
        """
        Parse a recipe string and add it to the recipes dict, and add its
        params to self.params

        A recipe is a string with one or more parameter names between {}
        indicating parts of an input value (a filepath, for example) which we want
        to capture into those parameters. For example:

        "{SURNAME}-{GIVENNAME}-{YYYY}{MM}{DD}.txt"

        when matched against

        "Duck-Donald-19300831.txt"

        should return the dict

        {
                "SURNAME": "Duck",
                "GIVENNAME": "Donald",
                "YYYY":
                "1930",
                "MM": "08",
                "DD": "31"
        }

        Converts a recipe into a regexp object with named group
        parameters. This can be used to match target strings and return the
        desired dict with the match.groupdict() function

        Parameter names which consist of one or more of the same character are
        converted into patterns which match that number of digits. The
        parameter 'ID' matches a list of one or more consecutive digits.

        All other parameter names are converted into a non-greedy match up to
        the next character after the parameter in the pattern, or the end of
        the recipe.

        "*" and "**" are special recipes for matching and ignoring one or more
        than one intervening directories. They are returned as-is without being
        converted to a regexp.

        A list of all the params is returned so that calling code can know what to
        expect from the pattern without having to run it or reparse the recipe.
        Params are listed in the order they appear in the pattern.

        ---
        recipe (str): a recipe as described abov

        Returns: ( list of str, re.Pattern )

        Raises: an re.error if for some reason the resulting regexp can't compile
        - for instance, if a parameter name has illegal characters or is repeated.
        """

        regexp = ""
        params = []
        if recipe == "*" or recipe == "**":
            return [], recipe
        for macro in re.finditer(r"{(.*?)}([^{]*)", recipe):
            param, delimiter = macro.group(1, 2)
            if param in params:
                raise RecipeException(f"Recipe has repeated parameter: {param}")
            if re.match(param[0] + "+$", param):
                regexp += f"(?P<{param}>" + (r"\d" * len(param)) + ")"
            # elif param == "ID":
            #     regexp += f"(?P<{param}>" + r"\d+" + ")"  # for Sublime Text
            else:
                if delimiter != "":
                    c = delimiter[0]
                    regexp += f"(?P<{param}>[^{c}]*?)"
                else:
                    regexp += f"(?P<{param}>.*)"
            regexp += re.escape(delimiter)
            params.append(param)
        return params, re.compile(regexp)

    def match(self, root, filepath):
        """
        Calls match_path to get values from the filepath, and then tries to
        read DICOM values from the file itself. Returns a FileMatch object
        whether matching succeeds or fails - the success flag and values will
        be set if it succeeded.
        ---
        file: pathlib.Path

        returns: a FileMatch
        """
        label, values = self.match_path(filepath.relative_to(root))
        if label:
            logger.debug(f"> reading DICOM metadata {filepath}")
            dicom_values = self.read_dicom(filepath)
            if dicom_values is None:
                logger.debug(f"> DICOM read failed {filepath}")
                match = FileMatch(self, filepath, None)
                match.status = "No DICOM metadata"
                return match
            return self.make_filematch(filepath, label, values, dicom_values)
        match = FileMatch(self, filepath, None)
        match.status = "unmatched"
        return match

    def make_filematch(self, file, label, path_values, dicom_values):
        """
        Map the values captured to XNAT hierarchy values and return an
        FileMatch object. If the mapping was unsuccessful, the FileMatch will
        have its success flag switched off
        """
        match = FileMatch(self, file, label)
        match.values = path_values
        match.dicom_values = dicom_values
        try:
            logger.debug("> mapping DICOM and path values to XNAT params")
            match.xnat_params = self.map_values(path_values, dicom_values)
            match.success = True
        except ValueError as e:
            logger.warning(f"> xnat params error: {e}")
            match.success = False
            match.status = "unmatched"
        match.selected = match.success
        return match

    def from_spreadsheet(self, row):
        """
        Build a FileMatch from a spreadsheet row, using the FileMatch.from_row
        method
        """
        match = FileMatch(self, None, None)
        match.from_row(row)
        return match

    def match_path(self, filepath):
        """
        Try to match a filepath against each of the recipes and return the label
        and values for the first one which matches.
        ---
        file: pathlib.Path

        returns: { str: str }
        """
        logger.debug(f"Trying to match {filepath}")
        for label, recipes in self.recipes.items():
            logger.debug(f"> pattern: {label}")
            values = self.match_recipe(recipes, filepath)
            if values:
                logger.debug(f"> successful path match for {label}")
                return label, values
        return None, None

    def match_recipe(self, patterns, path):
        """
        Matches a list of patterns against a pathlib.Path, returning None if
        no match was possible, or a dict of captured values.

        Matching starts from the beginning and works down the subdirectories
        by the following rules:

        - if the next subdir matches the next pattern, capture any values and
          move to the next subdir and pattern

        - if the pattern is "*", match it against any subdir and keep going

        - if the pattern is "**", match it against one or more subdirs until
          the rest of the path matches the rest of the patterns (a
          "non-greedy" match in regexp terms)

        Both "*" and "**" must match against at least one subdir.

        ---
        patterns: list of re.Pattern
        path: pathlib.Path

        Returns: None, or dict of { str: str }
        """
        dirs = list(path.parts)
        matchpatterns = patterns[:]
        values = self.match_recipe_r(matchpatterns, dirs)
        return values

    def match_recipe_r(self, patterns, dirs):
        """The recursive part of match_recipe. Checks if the head of patterns
        matches the head of dirs, returning None if there's no match or either
        patterns or dirs is exhausted early, and adds any captured values to
        values captured further down the list. Uses lookahead on '**' to see
        if the rest of the pattern after this dir matches. Won't be efficient
        on very deep hierarchies because of this.

        patterns: list of re.Pattern
        path: list of str

        Returns: None, or dict of { str: str }
        """

        if not patterns:
            if dirs:
                logger.debug(">> ran out of patterns before end of path")
                return None
            else:
                return {}  # reached the end of both at the same time
        if not dirs:
            logger.debug(">> ran out of path before end of patterns")
            return None
        values = {}
        tail_values = self.match_recipe_r(patterns[1:], dirs[1:])
        if patterns[0] == "**":
            logger.debug(f">> pattern ** / path {dirs[0]}")
            if tail_values is not None:
                logger.debug(">> Matched remainder of path")
                return tail_values
            logger.debug(">> moving to next path")
            return self.match_recipe_r(patterns, dirs[1:])
        if patterns[0] == "*":
            logger.debug(f">> pattern * / path {dirs[0]}")
            return tail_values
        logger.debug(f">> pattern {patterns[0]} / path {dirs[0]}")
        m = patterns[0].match(dirs[0])
        if not m:
            logger.debug(">> No match")
            return None
        values = m.groupdict()
        if tail_values is None:
            logger.debug(">> Match but no values")
            return None
        for k, v in tail_values.items():
            values[k] = v
            logger.debug(f">> Matched values {values}")
        return values

    def match_paths(self, patterns, dirs):
        """
        Given a list of patterns and path parts, this tries to match
        the first pattern with the first path. If there's only one
        pattern, it adds the extra condition that there must be only one
        path - this is so that the last pattern in a recipe will only
        match a file, not a directory
        """
        m = patterns[0].match(dirs[0])
        if m is not None:
            if len(patterns) > 1:
                return m
            if len(dirs) == 1:
                return m
        return None

    def read_dicom(self, file):
        """
        Try to parse the dicom metadata from the file and returns a dict
        of all the values extracted for this Matcher's dicom_params
        ---
        file: a pathlib.Path

        returns: { str: str }
        """
        try:
            dc_meta = dcmread(file)
            values = {p: dc_meta.get(p) for p in self._dicom_params}
            return values
        except InvalidDicomError:
            return None

    def map_values(self, path_values, dicom_values):
        """
        Given a dict of values which has been captured from a filepath by a recipe,
        try to map it to the XNAT hierarchy.

        Raises: ValueError if any of the required values are missing

        values: dict of { str: str }
        """
        for k, v in dicom_values.items():
            path_values["DICOM:" + k] = v
        xnat_params = {}
        for xnat_cat, path_vars in self.mappings.items():
            for v in path_vars:
                if v not in path_values:
                    raise ValueError(f"value {v} not found")
                if path_values[v] is None:
                    raise ValueError(f"value {v} is None")
            xnat_params[xnat_cat] = "".join([path_values[v] for v in path_vars])
            xnat_params[xnat_cat] = xnat_params[xnat_cat].replace(" ", "_")
        return xnat_params


class FileMatch:
    """
    Represents a file, which may or may not have been successfully matched.
    Used as the return value from Matcher.match and as the result of loading
    the log spreadsheet.
    """

    def __init__(self, matcher, file, label):
        self.matcher = matcher
        self.label = label
        self.file = str(file)
        self.filename = None
        if file is not None:
            self.filename = file.name
        self.values = None
        self.dicom_values = None
        self.xnat_params = None
        self.session_label = None
        self.error = None
        self.success = False
        self.status = None
        self.selected = None
        self._columns = None
        self._status = None
        self._selected = None

    @property
    def columns(self, refresh=False):
        """
        Returns this file's representation in the spreadsheet, which may or
        may not be a successful match
        """
        if self._columns is not None:
            return self._columns
        self._columns = [self.label, self.file, self.filename]
        if self.selected:
            self._columns += ["Y"]
        else:
            self._columns += ["N"]
        if self.error is not None:
            self._columns += [self.error]
        else:
            if self.status is None:
                self._columns += [""]
            else:
                self._columns += [self.status]
        self._columns += [self.session_label]
        self._columns += self.dict_to_columns(XNAT_HIERARCHY, self.xnat_params)
        self._columns += self.dict_to_columns(
            self.matcher.dicom_params, self.dicom_values
        )
        self._columns += self.dict_to_columns(self.matcher.params, self.values)
        return self._columns

    def dict_to_columns(self, columns, values):
        if values is None:
            return ["" for _ in columns]
        else:
            return [values.get(c, "") for c in columns]

    def from_row(self, row):
        """
        Read a row from the spreadsheet and populate the values required to
        upload.
        """
        self.label = row[0]
        self.file = row[1]
        self.filename = row[2]
        self.selected = row[3] == "Y"
        self.status = row[4]
        self.session_label = row[5]
        self.xnat_params = {
            "Subject": row[6],
            "Session": row[7],
            "Dataset": row[8],
        }
        self.dicom_values = {}
        c = 9
        for p in self.matcher.dicom_params:
            self.dicom_values[p] = row[c]
            c += 1
        self.values = {}
        for p in self.matcher.params:
            self.values[p] = row[c]
            c += 1

    def load_dicom(self):
        """
        Utility method used to load the dicom metadata for an umatched file
        when debugging
        """
        dicom_values = self.matcher.read_dicom(self.file)
        if dicom_values is not None:
            self.dicom_values = dicom_values
            self._columns = None

    @property
    def subject(self):
        if self.xnat_params is not None:
            return self.xnat_params["Subject"]
        else:
            return None

    @property
    def session(self):
        if self.xnat_params is not None:
            return self.xnat_params["Session"]
        else:
            return None

    @property
    def dataset(self):
        if self.xnat_params is not None:
            return self.xnat_params["Dataset"]
        else:
            return None

    @property
    def study_date(self):
        if self.session is not None:
            return self.session
        if self.dicom_values is not None:
            return self.dicom_values["StudyDate"]
        return ""

    @property
    def modality(self):
        if self.dicom_values is not None:
            return self.dicom_values["Modality"]
        else:
            return "OT"  # DICOM code for "Other"



from pathlib import Path
import random
import string
from datetime import datetime, timedelta
import logging

logger = logging.getLogger(__name__)

# Note: this test is a little clunky because Matcher.match now does DICOM
# lookups, so it requires a file to work on. These tests are just about
# path value extraction, which is why we're testing match_path.


def test_match(matcher_case):
    matcher, case = matcher_case
    for path in case["paths"]:
        _, results = matcher.match_path(Path(path["path"]))
        assert results == path["values"]


def test_no_match(matcher_case):
    matcher, case = matcher_case
    if "bad_paths" in case:
        for bad_path in case["bad_paths"]:
            _, results = matcher.match_path(Path(bad_path))
            assert is_incomplete(results)  #


def test_random(matcher_case):
    matcher, case = matcher_case
    for i in range(100):
        path, expect = make_random_path(case["patterns"])
        _, results = matcher.match_path(Path(path))
        assert results == expect


def random_word():
    n = random.randint(4, 20)
    return "".join([random.choice(string.ascii_letters) for i in range(n)])


def random_id():
    n = random.randint(1, 8)
    return "".join([random.choice(string.digits + "-_") for i in range(n)])


def random_date():
    dt = datetime.now() - timedelta(days=random.randint(1, 10000))
    return dt.strftime("%Y%m%d")


def pattern_to_path(pattern, subject, subjectid, session, filename):
    """
    Given a matcher pattern and a subject, session and filename, return
    a path which that pattern should match. This assumes that the pattern
    contains {name}-{ID}, {DDDDDDDD} and {Filename} and that these mappings
    will be used in the Matcher.
    """
    dirs = []
    for part in pattern:
        if part == "*":
            dirs.append(random_word())
        elif part == "**":
            dirs.extend([random_word() for i in range(random.randint(1, 10))])
        elif part == "{SubjectName}-{ID}":
            dirs.append(subject + "-" + subjectid)
        elif part == "{DDDDDDDD}":
            dirs.append(session)
        elif part == "{Filename}":
            dirs.append(filename)
    return "/".join(dirs)


def make_random_path(pattern):
    """
    Given a pattern, generates a random path which it will match, and
    the values which should be matched from it, expanding wildcards as
    necessary.
    """
    subject = random_word()
    subjectid = random_id()
    session = random_date()
    filename = random_word() + ".dcm"
    path = pattern_to_path(pattern, subject, subjectid, session, filename)
    return path, {
        "SubjectName": subject,
        "ID": subjectid,
        "DDDDDDDD": session,
        "Filename": filename,
    }


def is_incomplete(values):
    """
    A hack because match_path doesn't always return None on a bad match
    """
    if values is None:
        return True
    if "name" not in values:
        return True
    if "ID" not in values:
        return True
    if "DDDDDDDD" not in values:
        return True
    if "Filename" not in values:
        return True



# import xnatutils
import xnatuploader.put
import os.path
import logging

logger = logging.getLogger(__name__)


class Upload:
    """
    An Upload represents a session for a single patient and visit, which may
    have more than one file.
    """

    def __init__(self, session_label, subject, date, modality, scan_type):
        self.session_label = session_label
        self.subject = subject
        self.modality = modality
        self.date = date
        self.scan_type = scan_type
        self.new_session = True
        self.xnat_session = None
        self.files = []

    @property
    def label(self):
        """A value which can identify this upload in XNAT - scan_type
        corresponds to dataset. Used for logging upload errors which are
        triggered by start_upload below."""
        return f"{self.session_label}/{self.scan_type}"

    def add_file(self, file):
        self.files.append(file)

    def start_upload(self, xnat_session, project):
        """Create a resource in the session for this scan"""
        self.xnat_session = xnat_session
        self.resource = xnatuploader.put.resource(
            self.session_label,
            self.scan_type,
            resource_name="DICOM",
            project_id=project,
            subject_id=self.subject,
            date=self.date,
            modality=self.modality,
            create_session=self.new_session,
            connection=xnat_session,
        )

    def upload(self, files, overwrite=False):
        """
        Upload a batch of files to the current resource and checks their
        digests

        Args:
            files: list of Matchfile
            overwrite: boolean
        Returns:
            dict of { str: str } with a status message, "success" or an error
        ---
        """
        for file in files:
            fname = os.path.basename(file.file)
            if fname in self.resource.files:
                if overwrite:
                    self.resource.files[fname].delete()  # I am not sure if this is good
            self.resource.upload(file.file, fname)
        return self.check_digests(files)

    def check_digests(self, files):
        """Check the digests of a batch of files, and returns a hash-by-filename
        of success or failure
        """
        result = self.xnat_session.get(self.resource.uri + "/files")
        if result.status_code != 200:
            logger.error(
                f"Request for digests at {self.resource.uri} returned status {result.status_code}"
            )
            digests = {}
        else:
            digests = {
                f["Name"]: f["digest"] for f in result.json()["ResultSet"]["Result"]
            }
        status = {}
        for file in files:
            xnat_filename = os.path.basename(file.file)
            if xnat_filename not in digests:
                status[
                    file.file
                ] = f"File {file.file} {xnat_filename} not found in digests"
                logger.error(status[file.file])
                logger.error(digests)
            else:
                remote_digest = digests[xnat_filename]
                local_digest = xnatuploader.put.calculate_checksum(file.file)
                if local_digest != remote_digest:
                    status[
                        file.file
                    ] = f"Digest mismatch {local_digest} {remote_digest}"
                else:
                    status[file.file] = "success"
        return status

    def log(self, logger):
        """
        Write an upload batch to logger for debugging
        ---
        upload: a dict as returned by collate_upload
        """
        logger.info(f"Session: {self.session_label}")
        logger.info(f"    Subject: {self.upload}")
        logger.info(f"    Modality: {self.modality}")
        logger.info(f"    Scan type: {self.scan_type}")
        for file in self.files:
            logger.info(f"        File: {file.file}")
